"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var graphql_transformer_core_1 = require("graphql-transformer-core");
var resources_1 = require("./resources");
var graphql_transformer_common_1 = require("graphql-transformer-common");
var graphql_1 = require("graphql");
var OWNER_AUTH_STRATEGY = "owner";
var DEFAULT_OWNER_FIELD = "owner";
var DEFAULT_IDENTITY_FIELD = "username";
var GROUPS_AUTH_STRATEGY = "groups";
/**
 * Implements the ModelAuthTransformer.
 *
 * Owner Auth Usage:
 *
 * type Post @auth(allow: owner) {
 *   id: ID!
 *   title: String
 *   createdAt: String
 *   updatedAt: String
 * }
 *
 * Impact:
 *
 * getPost - In the response mapping template we check the "owner" field === $ctx.identity.username.
 * listPost - In the response mapping template we return only items where "owner" === $ctx.identity.username
 * createPost - We automatically insert a "owner" field to attribute values where "owner" === $ctx.identity.username.
 * updatePost - Expose "owner" field in input/output and would set conditional update expression to look for owner.
 * deletePost - Conditional expression checking that the owner === $ctx.identity.username
 *
 * Note: The name of the "owner" field may be configured via the CF paramaters.
 *
 * type Post @auth(allow: groups, groups: ["Admin", "Dev"]) {
 *   id: ID!
 *   title: String
 *   createdAt: String
 *   updatedAt: String
 * }
 *
 * Impact:
 *
 * getPost - Update req template to look for the groups in the identity.
 * listPost - Update req template to look for the groups in the identity.
 * createPost - Update req template to look for the groups in the identity.
 * updatePost - Update req template to look for the groups in the identity.
 * deletePost - Update req template to look for the groups in the identity.
 *
 * TODO: Document support for dynamic group authorization against
 * attributes of the records using conditional expressions. This will likely
 * be via a new argument such as "groupsField".
 */
var ModelAuthTransformer = /** @class */ (function (_super) {
    __extends(ModelAuthTransformer, _super);
    function ModelAuthTransformer() {
        var _this = _super.call(this, 'ModelAuthTransformer', "\n            directive @auth(rules: [AuthRule!]!) on OBJECT\n            input AuthRule {\n                allow: AuthStrategy!\n                ownerField: String # defaults to \"owner\"\n                identityField: String # defaults to \"username\"\n                groupsField: String\n                groups: [String]\n                queries: [ModelQuery]\n                mutations: [ModelMutation]\n            }\n            enum AuthStrategy { owner groups }\n            enum ModelQuery { get list }\n            enum ModelMutation { create update delete }\n            ") || this;
        /**
         * Updates the GraphQL API record to use user pool auth.
         */
        _this.updateAPIForUserPools = function (ctx) {
            var apiRecord = ctx.getResource(graphql_transformer_common_1.ResourceConstants.RESOURCES.GraphQLAPILogicalID);
            var updated = _this.resources.updateGraphQLAPIWithAuth(apiRecord);
            ctx.setResource(graphql_transformer_common_1.ResourceConstants.RESOURCES.GraphQLAPILogicalID, updated);
        };
        _this.before = function (ctx) {
            var template = _this.resources.initTemplate();
            ctx.mergeResources(template.Resources);
            ctx.mergeParameters(template.Parameters);
            ctx.mergeOutputs(template.Outputs);
            ctx.mergeConditions(template.Conditions);
            _this.updateAPIForUserPools(ctx);
        };
        /**
         * Implement the transform for an object type. Depending on which operations are to be protected
         */
        _this.object = function (def, directive, ctx) {
            var get = function (s) { return function (arg) { return arg.name.value === s; }; };
            var getArg = function (arg, dflt) {
                var argument = directive.arguments.find(get(arg));
                return argument ? graphql_1.valueFromASTUntyped(argument.value) : dflt;
            };
            var modelDirective = def.directives.find(function (dir) { return dir.name.value === 'model'; });
            if (!modelDirective) {
                throw new graphql_transformer_core_1.InvalidDirectiveError('Types annotated with @auth must also be annotated with @model.');
            }
            // Get the auth rules.
            var rules = getArg('rules', []);
            // Create a reverse index on rules from operation -> rules list.
            var queryRules = {
                get: [],
                list: [],
            };
            var mutationRules = {
                create: [],
                update: [],
                delete: []
            };
            var matchQuery = function (op) { return function (rule) {
                if (rule.queries) {
                    var matchesOp = rule.queries.find(function (o) { return o === op; });
                    return Boolean(matchesOp);
                }
                else if (rule.queries === null) {
                    return false;
                }
                return true;
            }; };
            var matchMutation = function (op) { return function (rule) {
                if (rule.mutations) {
                    var matchesOp = rule.mutations.find(function (o) { return o === op; });
                    return Boolean(matchesOp);
                }
                else if (rule.mutations === null) {
                    return false;
                }
                return true;
            }; };
            for (var _i = 0, rules_1 = rules; _i < rules_1.length; _i++) {
                var rule = rules_1[_i];
                if (matchQuery('get')(rule)) {
                    queryRules.get.push(rule);
                }
                if (matchQuery('list')(rule)) {
                    queryRules.list.push(rule);
                }
                if (matchMutation('create')(rule)) {
                    mutationRules.create.push(rule);
                }
                if (matchMutation('update')(rule)) {
                    mutationRules.update.push(rule);
                }
                if (matchMutation('delete')(rule)) {
                    mutationRules.delete.push(rule);
                }
            }
            // For each operation evaluate the rules and apply the changes to the relevant resolver.
            _this.protectCreateMutation(ctx, graphql_transformer_common_1.ResolverResourceIDs.DynamoDBCreateResolverResourceID(def.name.value), mutationRules.create);
            _this.protectUpdateMutation(ctx, graphql_transformer_common_1.ResolverResourceIDs.DynamoDBUpdateResolverResourceID(def.name.value), mutationRules.update);
            _this.protectDeleteMutation(ctx, graphql_transformer_common_1.ResolverResourceIDs.DynamoDBDeleteResolverResourceID(def.name.value), mutationRules.delete);
            _this.protectGetQuery(ctx, graphql_transformer_common_1.ResolverResourceIDs.DynamoDBGetResolverResourceID(def.name.value), queryRules.get);
            _this.protectListQuery(ctx, graphql_transformer_common_1.ResolverResourceIDs.DynamoDBListResolverResourceID(def.name.value), queryRules.list);
        };
        _this.resources = new resources_1.ResourceFactory();
        return _this;
    }
    /**
     * Get queries are protected by prepending logic to the response mapping template
     * that validates $ctx.result against the specified auth rules.
     * @param ctx The transformer context.
     * @param resolverResourceId The logical id of the get resolver.
     * @param rules The auth rules to apply.
     */
    ModelAuthTransformer.prototype.protectGetQuery = function (ctx, resolverResourceId, rules) {
        var resolver = ctx.getResource(resolverResourceId);
        if (!rules || rules.length === 0 || !resolver) {
            return;
        }
        var beforeResponse = [];
        for (var _i = 0, rules_2 = rules; _i < rules_2.length; _i++) {
            var rule = rules_2[_i];
            if (rule.allow === OWNER_AUTH_STRATEGY) {
                var ownerField = rule.ownerField || OWNER_AUTH_STRATEGY;
                var identityField = rule.identityField || DEFAULT_IDENTITY_FIELD;
                var authSnippet = this.resources.ownerGetResolverResponseMappingTemplateSnippet(ownerField, identityField);
                beforeResponse.push(authSnippet);
            }
            else if (rule.allow === GROUPS_AUTH_STRATEGY) {
                if (rule.groups) {
                    var authSnippet = this.resources.staticGroupAuthorizationResponseMappingTemplate(rule.groups);
                    beforeResponse.push(authSnippet);
                }
                else if (rule.groupsField) {
                    var authSnippet = this.resources.dynamicGroupGetResolverResponseMappingTemplateSnippet(rule.groupsField);
                    beforeResponse.push(authSnippet);
                }
                else {
                    throw new graphql_transformer_core_1.InvalidDirectiveError("@auth(allow: groups ...) must also be passed \"groups\" or \"groupsField\".");
                }
            }
        }
        // Join together the authorization expressions and update the resolver.
        var templateParts = beforeResponse.concat([this.resources.throwWhenUnauthorized(), resolver.Properties.ResponseMappingTemplate]);
        resolver.Properties.ResponseMappingTemplate = templateParts.join('\n\n');
        ctx.setResource(resolverResourceId, resolver);
    };
    /**
     * List queries are protected by prepending logic to the response mapping template
     * that filters $ctx.result.items for items that satisfy the auth rules.
     * This is semi-involved and the logic depends on what @auth config is being used.
     * @param ctx The transformer context.
     * @param resolverResourceId The logical id of the resolver to be updated in the CF template.
     * @param rules The set of rules that apply to the operation.
     */
    ModelAuthTransformer.prototype.protectListQuery = function (ctx, resolverResourceId, rules) {
        var resolver = ctx.getResource(resolverResourceId);
        if (!rules || rules.length === 0 || !resolver) {
            return;
        }
        var beforeExpressions = [];
        var itemEquivalenceExpressions = [];
        var beforeItemEquivalenceExpression = [];
        for (var _i = 0, rules_3 = rules; _i < rules_3.length; _i++) {
            var rule = rules_3[_i];
            if (rule.allow === OWNER_AUTH_STRATEGY) {
                var ownerField = rule.ownerField || OWNER_AUTH_STRATEGY;
                var identityField = rule.identityField || DEFAULT_IDENTITY_FIELD;
                var authSnippet = this.resources.ownerListResolverItemCheck(ownerField, identityField);
                itemEquivalenceExpressions.push(authSnippet);
            }
            else if (rule.allow === GROUPS_AUTH_STRATEGY) {
                if (rule.groups) {
                    var authSnippet = this.resources.staticGroupAuthorizationResponseMappingTemplateAST(rule.groups);
                    // At the beginning of the resolver we run the static group check. If isAuthorized is set to true
                    // then allow every item to be appended.
                    beforeExpressions.push(authSnippet);
                    itemEquivalenceExpressions.push(this.resources.isAuthorized());
                }
                else if (rule.groupsField) {
                    var authSnippet = this.resources.dynamicGroupListBeforeItemEquivalenceExpressionAST(rule.groupsField);
                    // Run some logic within the loop before the item equivalence expression to compare the fetched values
                    // groupsField against the logged in identity and updating $isAuthorized appropriately.
                    beforeItemEquivalenceExpression.push(authSnippet);
                    itemEquivalenceExpressions.push(this.resources.isAuthorizedLocallyOrGlobally());
                }
                else {
                    throw new graphql_transformer_core_1.InvalidDirectiveError("@auth(allow: groups ...) must also be passed \"groups\" or \"groupsField\".");
                }
            }
        }
        // Join together the authorization expressions and update the resolver.
        var templateParts = [
            this.resources.loopThroughResultItemsAppendingAuthorized(itemEquivalenceExpressions, beforeExpressions, beforeItemEquivalenceExpression),
            resolver.Properties.ResponseMappingTemplate
        ];
        resolver.Properties.ResponseMappingTemplate = templateParts.join('\n\n');
        ctx.setResource(resolverResourceId, resolver);
    };
    /**
     * In create mutations the $ctx.identity.username is automatically
     * injected into the object.
     * @param ctx
     * @param resolverResourceId
     * @param rules
     */
    ModelAuthTransformer.prototype.protectCreateMutation = function (ctx, resolverResourceId, rules) {
        var resolver = ctx.getResource(resolverResourceId);
        if (!rules || rules.length === 0 || !resolver) {
            return;
        }
        var beforeRequest = [];
        for (var _i = 0, rules_4 = rules; _i < rules_4.length; _i++) {
            var rule = rules_4[_i];
            if (rule.allow === OWNER_AUTH_STRATEGY) {
                var ownerField = rule.ownerField || DEFAULT_OWNER_FIELD;
                var identityField = rule.identityField || DEFAULT_IDENTITY_FIELD;
                var authSnippet = this.resources.ownerCreateResolverRequestMappingTemplateSnippet(ownerField, identityField);
                beforeRequest.push(authSnippet);
            }
            else if (rule.allow === GROUPS_AUTH_STRATEGY) {
                if (rule.groups) {
                    var authSnippet = this.resources.staticGroupAuthorizationResponseMappingTemplate(rule.groups);
                    beforeRequest.push(authSnippet);
                    beforeRequest.push(this.resources.throwWhenUnauthorized());
                }
                else if (rule.groupsField) {
                    var authSnippet = this.resources.dynamicGroupCreateResolverRequestMappingTemplateSnippet(rule.groupsField);
                    beforeRequest.push(authSnippet);
                    beforeRequest.push(this.resources.throwWhenUnauthorized());
                }
                else {
                    throw new graphql_transformer_core_1.InvalidDirectiveError("@auth(allow: groups ...) must also be passed \"groups\" or \"groupsField\".");
                }
            }
        }
        // Join together the authorization expressions and update the resolver.
        var templateParts = beforeRequest.concat([resolver.Properties.RequestMappingTemplate]);
        resolver.Properties.RequestMappingTemplate = templateParts.join('\n\n');
        ctx.setResource(resolverResourceId, resolver);
    };
    /**
     * Both protect and update mutations operate via an injected condition expression
     * in the DynamoDB request. The DynamoDBTransformer knows to look for a special
     * auth condition in the template and will merge in values accordingly if the user
     * has also supplied their own filter expressions.
     * @param ctx The transformer context.
     * @param resolverResourceId The logical id of the resolver in the template.
     * @param rules The list of rules to apply.
     */
    ModelAuthTransformer.prototype.protectUpdateOrDeleteMutation = function (ctx, resolverResourceId, rules) {
        var resolver = ctx.getResource(resolverResourceId);
        if (!rules || rules.length === 0 || !resolver) {
            return;
        }
        var beforeRequest = [];
        for (var _i = 0, rules_5 = rules; _i < rules_5.length; _i++) {
            var rule = rules_5[_i];
            if (rule.allow === OWNER_AUTH_STRATEGY) {
                var ownerField = rule.ownerField || DEFAULT_OWNER_FIELD;
                var identityField = rule.identityField || DEFAULT_IDENTITY_FIELD;
                var authSnippet = this.resources.ownerUpdateAndDeleteResolverRequestMappingTemplateSnippet(ownerField, identityField);
                beforeRequest.push(authSnippet);
            }
            else if (rule.allow === GROUPS_AUTH_STRATEGY) {
                if (rule.groups) {
                    var authSnippet = this.resources.staticGroupAuthorizationResponseMappingTemplate(rule.groups);
                    beforeRequest.push(authSnippet);
                    beforeRequest.push(this.resources.throwWhenUnauthorized());
                }
                else if (rule.groupsField) {
                    var authSnippet = this.resources.dynamicGroupUpdateAndDeleteResolverRequestMappingTemplateSnippet(rule.groupsField);
                    beforeRequest.push(authSnippet);
                }
                else {
                    throw new graphql_transformer_core_1.InvalidDirectiveError("@auth(allow: groups ...) must also be passed \"groups\" or \"groupsField\".");
                }
            }
        }
        // Join together the authorization expressions and update the resolver.
        var templateParts = beforeRequest.concat([resolver.Properties.RequestMappingTemplate]);
        resolver.Properties.RequestMappingTemplate = templateParts.join('\n\n');
        ctx.setResource(resolverResourceId, resolver);
    };
    ModelAuthTransformer.prototype.protectUpdateMutation = function (ctx, resolverResourceId, rules) {
        return this.protectUpdateOrDeleteMutation(ctx, resolverResourceId, rules);
    };
    ModelAuthTransformer.prototype.protectDeleteMutation = function (ctx, resolverResourceId, rules) {
        return this.protectUpdateOrDeleteMutation(ctx, resolverResourceId, rules);
    };
    return ModelAuthTransformer;
}(graphql_transformer_core_1.Transformer));
exports.ModelAuthTransformer = ModelAuthTransformer;
//# sourceMappingURL=ModelAuthTransformer.js.map